package edu.kit.kastel.sdq.accesscontrolgenerator.validation;

import edu.kit.kastel.sdq.AccessControlMetamodel.AccessControlSystem.AccessControlSystem;
import edu.kit.kastel.sdq.AccessControlMetamodel.AccessControlSystem.AccessControlSystemFactory;
import edu.kit.kastel.sdq.AccessControlMetamodel.AccessControlSystem.FunctionToFunctionRelation;
import edu.kit.kastel.sdq.AccessControlMetamodel.AccessControlSystem.Role;
import edu.kit.kastel.sdq.AccessControlMetamodel.AccessControlSystem.VariableToVariableRelation;
import java.util.ArrayList;
import java.util.Collection;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;

/**
 * Calculates the transitive closures for the AccessControlSystem to support the later reasoning about
 * possible violations.
 */
@SuppressWarnings("all")
public class TransitiveClosureCalculator {
  private final int MIN_AMOUNT_FOR_TRANS_CLOSURE = 2;
  
  private final int MIN_AMOUNT_FOR_TRANS_ROLE_HIERARCHY = 3;
  
  private AccessControlSystemFactory factory = AccessControlSystemFactory.eINSTANCE;
  
  /**
   * Calculates the transitive closures for the influence relation between variables and the call relation between functions.
   * Both of these are necessary for a complete reasoning about the systems soundness.
   * If less then 2 relation tuples exist, nothing is done. Otherwise the transitive tuples are calculated recursively.
   */
  public void calculateTransitiveClosures(final AccessControlSystem system) {
    int _length = ((Object[])Conversions.unwrapArray(system.getVariableToVariableTuples(), Object.class)).length;
    boolean _greaterEqualsThan = (_length >= this.MIN_AMOUNT_FOR_TRANS_CLOSURE);
    if (_greaterEqualsThan) {
      final Collection<VariableToVariableRelation> varClosure = this.calculateTransitiveClosureForVariableInfluence(system.getVariableToVariableTuples());
      system.getVariableToVariableTuples().clear();
      system.getVariableToVariableTuples().addAll(varClosure);
    }
    int _length_1 = ((Object[])Conversions.unwrapArray(system.getFunctionToFunctionTuples(), Object.class)).length;
    boolean _greaterEqualsThan_1 = (_length_1 >= this.MIN_AMOUNT_FOR_TRANS_CLOSURE);
    if (_greaterEqualsThan_1) {
      final Collection<FunctionToFunctionRelation> funcClosure = this.calculateTransitiveClosureForFunctionCalls(system.getFunctionToFunctionTuples());
      system.getFunctionToFunctionTuples().clear();
      system.getFunctionToFunctionTuples().addAll(funcClosure);
    }
    int _length_2 = ((Object[])Conversions.unwrapArray(system.getRoles(), Object.class)).length;
    boolean _greaterEqualsThan_2 = (_length_2 >= this.MIN_AMOUNT_FOR_TRANS_ROLE_HIERARCHY);
    if (_greaterEqualsThan_2) {
      EList<Role> _roles = system.getRoles();
      for (final Role role : _roles) {
        {
          final ArrayList<Role> alreadyConsidered = new ArrayList<Role>();
          alreadyConsidered.add(role);
          final Collection<Role> superiors = this.calculateTransitiveClosureForRoleHierarchy(role, alreadyConsidered);
          role.getSuperior().clear();
          role.getSuperior().addAll(superiors);
        }
      }
    }
  }
  
  /**
   * Recursively calculates the transitive closure for variable to variable relations.
   * In each call, all new relations are calculated and the recursion is stopped if no new relations are found.
   * For each relation, a new relation element is created using the AccessControlSystemFactory.
   */
  private Collection<VariableToVariableRelation> calculateTransitiveClosureForVariableInfluence(final Collection<VariableToVariableRelation> oldClosure) {
    boolean _isEmpty = oldClosure.isEmpty();
    if (_isEmpty) {
      return oldClosure;
    }
    final ArrayList<VariableToVariableRelation> newClosure = new ArrayList<VariableToVariableRelation>();
    newClosure.addAll(oldClosure);
    for (final VariableToVariableRelation varToVar : oldClosure) {
      {
        final Function1<VariableToVariableRelation, Boolean> _function = new Function1<VariableToVariableRelation, Boolean>() {
          public Boolean apply(final VariableToVariableRelation vv) {
            return Boolean.valueOf(vv.getInfluencer().equals(varToVar.getInfluenced()));
          }
        };
        final Iterable<VariableToVariableRelation> transitives = IterableExtensions.<VariableToVariableRelation>filter(oldClosure, _function);
        for (final VariableToVariableRelation trans : transitives) {
          final Function1<VariableToVariableRelation, Boolean> _function_1 = new Function1<VariableToVariableRelation, Boolean>() {
            public Boolean apply(final VariableToVariableRelation vv) {
              return Boolean.valueOf((vv.getInfluencer().equals(varToVar.getInfluencer()) && vv.getInfluenced().equals(trans.getInfluenced())));
            }
          };
          boolean _exists = IterableExtensions.<VariableToVariableRelation>exists(oldClosure, _function_1);
          boolean _not = (!_exists);
          if (_not) {
            final VariableToVariableRelation newVarToVar = this.factory.createVariableToVariableRelation();
            newVarToVar.setInfluencer(varToVar.getInfluencer());
            newVarToVar.setInfluenced(trans.getInfluenced());
            StringConcatenation _builder = new StringConcatenation();
            String _entityName = newVarToVar.getInfluencer().getEntityName();
            _builder.append(_entityName);
            _builder.append(" to ");
            String _entityName_1 = newVarToVar.getInfluenced().getEntityName();
            _builder.append(_entityName_1);
            newVarToVar.setEntityName(_builder.toString());
            newClosure.add(newVarToVar);
          }
        }
      }
    }
    int _length = ((Object[])Conversions.unwrapArray(newClosure, Object.class)).length;
    int _length_1 = ((Object[])Conversions.unwrapArray(oldClosure, Object.class)).length;
    boolean _equals = (_length == _length_1);
    if (_equals) {
      return newClosure;
    } else {
      return this.calculateTransitiveClosureForVariableInfluence(newClosure);
    }
  }
  
  /**
   * Recursively calculates the transitive closure for function to function relations.
   * In each call, all new relations are calculated and the recursion is stopped if no new relations are found.
   * For each relation, a new relation element is created using the AccessControlSystemFactory.
   */
  private Collection<FunctionToFunctionRelation> calculateTransitiveClosureForFunctionCalls(final Collection<FunctionToFunctionRelation> oldClosure) {
    boolean _isEmpty = oldClosure.isEmpty();
    if (_isEmpty) {
      return oldClosure;
    }
    final ArrayList<FunctionToFunctionRelation> newClosure = new ArrayList<FunctionToFunctionRelation>();
    newClosure.addAll(oldClosure);
    for (final FunctionToFunctionRelation funcToFunc : oldClosure) {
      {
        final Function1<FunctionToFunctionRelation, Boolean> _function = new Function1<FunctionToFunctionRelation, Boolean>() {
          public Boolean apply(final FunctionToFunctionRelation ff) {
            return Boolean.valueOf(ff.getCaller().equals(funcToFunc.getCallee()));
          }
        };
        final Iterable<FunctionToFunctionRelation> transitives = IterableExtensions.<FunctionToFunctionRelation>filter(oldClosure, _function);
        for (final FunctionToFunctionRelation trans : transitives) {
          final Function1<FunctionToFunctionRelation, Boolean> _function_1 = new Function1<FunctionToFunctionRelation, Boolean>() {
            public Boolean apply(final FunctionToFunctionRelation ff) {
              return Boolean.valueOf((ff.getCaller().equals(funcToFunc.getCaller()) && ff.getCallee().equals(trans.getCallee())));
            }
          };
          boolean _exists = IterableExtensions.<FunctionToFunctionRelation>exists(oldClosure, _function_1);
          boolean _not = (!_exists);
          if (_not) {
            final FunctionToFunctionRelation newFuncToFunc = this.factory.createFunctionToFunctionRelation();
            newFuncToFunc.setCaller(funcToFunc.getCaller());
            newFuncToFunc.setCallee(trans.getCallee());
            StringConcatenation _builder = new StringConcatenation();
            String _entityName = newFuncToFunc.getCaller().getEntityName();
            _builder.append(_entityName);
            _builder.append(" to ");
            String _entityName_1 = newFuncToFunc.getCallee().getEntityName();
            _builder.append(_entityName_1);
            newFuncToFunc.setEntityName(_builder.toString());
            newClosure.add(newFuncToFunc);
          }
        }
      }
    }
    int _length = ((Object[])Conversions.unwrapArray(newClosure, Object.class)).length;
    int _length_1 = ((Object[])Conversions.unwrapArray(oldClosure, Object.class)).length;
    boolean _equals = (_length == _length_1);
    if (_equals) {
      return newClosure;
    } else {
      return this.calculateTransitiveClosureForFunctionCalls(newClosure);
    }
  }
  
  /**
   * This function calculates the transitive closure for the role hierarchy by recursively
   * calling itself for all superiors found for the given base role. To keep track of the
   * already considered roles, a collection is carried along. If a role is considered, it is
   * added to that collection.
   */
  private Collection<Role> calculateTransitiveClosureForRoleHierarchy(final Role base, final Collection<Role> alreadyCons) {
    final ArrayList<Role> newClosure = new ArrayList<Role>();
    if ((base == null)) {
      return newClosure;
    }
    boolean _isEmpty = base.getSuperior().isEmpty();
    boolean _not = (!_isEmpty);
    if (_not) {
      EList<Role> _superior = base.getSuperior();
      for (final Role superior : _superior) {
        boolean _contains = alreadyCons.contains(superior);
        boolean _not_1 = (!_contains);
        if (_not_1) {
          alreadyCons.add(superior);
          newClosure.add(superior);
          newClosure.addAll(this.calculateTransitiveClosureForRoleHierarchy(superior, alreadyCons));
        }
      }
    }
    return newClosure;
  }
}
